from collections import namedtuple, defaultdict
from mammut.common.storage.storage_manager import StorageManager
from mammut.common.lexicon.feature import FeatureDescriptor
import mammut.common.lexicon.dictionary
from mammut.common.lexicon.pos import POSdescriptor
from mammut.common.lexicon.mammut_brat_configuration import (
    MammutBratConfiguration,
)
from mammut.common.lexicon.functional import FunctionalLemmaDescriptor
from mammut.common.lexicon.non_functional import NonFunctionalLemmaDescriptor
from mammut.common.lexicon.inflection_rule import InflectionRule
import pandas as pd

InVocabulary = namedtuple(
    "InVocabulary",
    [
        "value",
        "list",
        "description",
        "details",
        "base_value",
        "base_list",
        "base_description",
    ],
)


class LinguisticStandard:
    REGIONAL_SETTINGS_LEMMA_LIST_COLUMN_NAME = "regional-settings-lemma-list"
    REGIONAL_SETTINGS_DEFINITION_LIST_COLUMN_NAME = (
        "regional-settings-definition-list"
    )
    WORD_TYPE_LIST_COLUMN_NAME = "word-type-list"

    FEATURE_CLASS_GROUP_COLUMN_NAME = "feature-class-group"
    FEATURE_CLASS_COLUMN_NAME = "feature-class"
    FEATURE_NAME_COLUMN_NAME = "feature-name"
    FEATURE_ID_COLUMN_NAME = "feature-id"
    POS_TAGS_COLUMN_NAME = "pos-tags"
    DOCUMENTATION_LINK_COLUMN_NAME = "documentation-link"
    STATISTICS_LINK_COLUMN_NAME = "statistics-link"

    SOURCE_ID_COLUMN_NAME = "source-id"
    SOURCE_URL_PATTERN_COLUMN_NAME = "source-url-pattern"
    HTML_WRAPPER_COLUMN_NAME = "html-wrapper"
    REGIONAL_SETTINGS_COLUMN_NAME = "regional-settings"

    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "M"

    def __init__(
        self,
        spreadsheet_id: str,
        base_sheet_title: str,
        dict_sources_sheet_title: str,
        pos_sheet_title: str,
        features_sheet_title: str,
        dimensiones_sheet_title: str,
        functionals_sheet_title: str,
        non_functionals_sheet_title: str,
        tokenizer,
        primes_sheet_title,
        synonym_sheet_title: str,
    ):
        self.valid = True
        self.errors = []
        self.spreadsheet_id = spreadsheet_id
        self.base_sheet_title = base_sheet_title
        self.dict_sources_sheet_title = dict_sources_sheet_title
        self.pos_sheet_title = pos_sheet_title
        self.features_sheet_title = features_sheet_title
        self.dimensiones_sheet_title = dimensiones_sheet_title
        self.functionals_sheet_title = functionals_sheet_title
        self.non_functionals_sheet_title = non_functionals_sheet_title
        self.tokenizer = tokenizer
        self.primes_sheet_title = primes_sheet_title
        self.synonym_sheet_title = synonym_sheet_title
        self.regional_settings_lemma_list = []
        self.regional_settings_definition_list = []
        self.word_type_list = []
        self.pos_descriptor_list = {}
        self.pos_descriptor_classes = defaultdict(list)
        self.pos_annotation_shifts = defaultdict(list)
        self.feature_descriptor_list = {}
        self.feature_descriptor_by_pos_tag = {}
        self.all_feature_descriptor_by_pos_tag = defaultdict(list)
        self.dimensiones_descriptor_list = {}
        self.morpheme_paradigms = {}
        self.misspelling_paradigms = {}
        self.feature_paradigms = {}
        self.external_dictionary_sources = {}
        # Contains FunctionalLemmaDescriptor indexed by lemma, regional setting and pos concatanated by "-"
        self.functionals = {}
        self.functionals_implicit = {}
        self.functionals_by_regional_settings_lemma = {}
        self.functionals_by_regional_settings_lemma_implicit = {}
        self.undefined_functionals = {}
        self.non_functionals = {}
        self.base_range_name = (
            self.base_sheet_title + "!" + LinguisticStandard.DEFAULT_START_CELL
        )
        self.pos_range_name = (
            self.pos_sheet_title + "!" + LinguisticStandard.DEFAULT_START_CELL
        )
        # Load dimensions of annotation.
        self.dimensiones_range_name = (
                self.dimensiones_sheet_title
                + "!"
                + LinguisticStandard.DEFAULT_START_CELL
        )
        self.storage_manager = StorageManager()
        self.load_base_dataframe()
        self.load_features_dataframe()
        self.load_dictionary_dataframe()
        self.load_pos_dataframe()
        self.load_dimensions_dataframe()
        self.load_functionals_dataframe()
        self.load_non_functionals_dataframe()
        self.working_dictionary = None

    def load_base_dataframe(self):
        self.base_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.base_sheet_title,
            self.DEFAULT_START_COLUMN,
            self.DEFAULT_START_ROW,
            self.DEFAULT_END_CELL,
        )

        if self.base_data_frame is not None:
            for rsl in self.base_data_frame[
                LinguisticStandard.REGIONAL_SETTINGS_LEMMA_LIST_COLUMN_NAME
            ]:
                self.regional_settings_lemma_list.append(rsl)
            for rsd in self.base_data_frame[
                LinguisticStandard.REGIONAL_SETTINGS_DEFINITION_LIST_COLUMN_NAME
            ]:
                self.regional_settings_definition_list.append(rsd)
            for wt in self.base_data_frame[
                LinguisticStandard.WORD_TYPE_LIST_COLUMN_NAME
            ]:
                self.word_type_list.append(wt)
        else:
            self.valid = False
            self.errors.append(f"The base sheet is not present.")
        self.features_range_name = (
                self.features_sheet_title
                + "!"
                + LinguisticStandard.DEFAULT_START_CELL
        )

    def load_features_dataframe(self):
        self.features_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.features_sheet_title,
            LinguisticStandard.DEFAULT_START_COLUMN,
            LinguisticStandard.DEFAULT_START_ROW,
            LinguisticStandard.DEFAULT_END_CELL,
        )

        if self.features_data_frame is not None:
            for r in self.features_data_frame.iterrows():
                fet_desc = FeatureDescriptor(
                    r[1][LinguisticStandard.FEATURE_CLASS_GROUP_COLUMN_NAME],
                    r[1][LinguisticStandard.FEATURE_CLASS_COLUMN_NAME],
                    r[1][LinguisticStandard.FEATURE_NAME_COLUMN_NAME],
                    r[1][LinguisticStandard.FEATURE_ID_COLUMN_NAME],
                    r[1][LinguisticStandard.POS_TAGS_COLUMN_NAME],
                    r[1][LinguisticStandard.DOCUMENTATION_LINK_COLUMN_NAME],
                    r[1][LinguisticStandard.STATISTICS_LINK_COLUMN_NAME],
                    self,
                )
                self.feature_descriptor_list[fet_desc.feature_id] = fet_desc
        else:
            self.valid = False
            self.errors.append(f"The features sheet is not present.")
        self.dict_sources_range_name = (
                self.dict_sources_sheet_title
                + "!"
                + LinguisticStandard.DEFAULT_START_CELL
        )

    def load_dictionary_dataframe(self):
        self.dict_sources_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.dict_sources_sheet_title,
            LinguisticStandard.DEFAULT_START_COLUMN,
            LinguisticStandard.DEFAULT_START_ROW,
            LinguisticStandard.DEFAULT_END_CELL,
        )

        if self.dict_sources_data_frame is not None:
            for r in self.dict_sources_data_frame.iterrows():
                ext_dic = mammut.common.lexicon.dictionary.ExternalDictionarySource(
                    r[1][LinguisticStandard.SOURCE_ID_COLUMN_NAME],
                    r[1][LinguisticStandard.SOURCE_URL_PATTERN_COLUMN_NAME],
                    r[1][LinguisticStandard.HTML_WRAPPER_COLUMN_NAME],
                    r[1][LinguisticStandard.REGIONAL_SETTINGS_COLUMN_NAME],
                    self,
                )
                self.external_dictionary_sources[ext_dic.source_id] = ext_dic
        else:
            self.valid = False
            self.errors.append(
                f"The dicttionary external sources sheet is not present."
            )

    def load_pos_dataframe(self):
        self.pos_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.pos_sheet_title,
            LinguisticStandard.DEFAULT_START_COLUMN,
            LinguisticStandard.DEFAULT_START_ROW,
            LinguisticStandard.DEFAULT_END_CELL,
        )

        if self.pos_data_frame is not None:
            for r in self.pos_data_frame.iterrows():
                pos_desc = POSdescriptor(
                    r[1][POSdescriptor.POS_CLASS_GROUP_COLUMN_NAME],
                    r[1][POSdescriptor.POS_CLASS_COLUMN_NAME],
                    r[1][POSdescriptor.POS_NAME_COLUMN_NAME],
                    r[1][POSdescriptor.POS_ID_COLUMN_NAME],
                    r[1][POSdescriptor.POS_TYPE_COLUMN_NAME],
                    r[1][POSdescriptor.MORPHEME_PARADIGMS_COLUMN_NAME],
                    r[1][POSdescriptor.MISSPELLING_PARADIGMS_COLUMN_NAME],
                    r[1][POSdescriptor.FEATURES_COLUMN_NAME],
                    r[1][POSdescriptor.REGIONAL_SETTINGS_COLUMN_NAME],
                    r[1][POSdescriptor.ARC_COLUMN_NAME],
                    r[1][POSdescriptor.POS_ABBREVIATION],
                    r[1][POSdescriptor.POS_SHIFTS],
                    self,
                )
                self.morpheme_paradigms[
                    pos_desc.id
                ] = pos_desc.morpheme_paradigms
                self.misspelling_paradigms[
                    pos_desc.id
                ] = pos_desc.misspelling_paradigms
                self.feature_paradigms[
                    pos_desc.id
                ] = pos_desc.feature_paradigms
                self.pos_descriptor_list[pos_desc.id] = pos_desc
                self.pos_descriptor_classes[pos_desc.pos_class].append(
                    pos_desc
                )
            for posd in self.pos_descriptor_list.values():
                posd.set_pos_references(self.pos_descriptor_list)
                self.pos_annotation_shifts[posd.id].append(posd.id)
            for posd in self.pos_descriptor_list.values():
                for pds_id in posd.pos_shifts_ids:
                    if posd.id not in self.pos_annotation_shifts[pds_id]:
                        self.pos_annotation_shifts[pds_id].append(posd.id)
        else:
            self.valid = False
            self.errors.append(f"The pos-tags sheet is not present.")
        for fd in self.feature_descriptor_list.values():
            for pt_id in fd.pos_tags:
                f_list = []
                if pt_id in self.feature_descriptor_by_pos_tag:
                    f_list = self.feature_descriptor_by_pos_tag[pt_id]
                else:
                    self.feature_descriptor_by_pos_tag[pt_id] = f_list
                pt = self.pos_descriptor_list[pt_id]
                for morpheme_paradigm in pt.morpheme_paradigms.values():
                    if (
                            fd.feature_class
                            not in morpheme_paradigm.feature_classes
                    ):
                        f_list.append(fd)
                self.all_feature_descriptor_by_pos_tag[pt_id].append(fd)

    def load_dimensions_dataframe(self):
        self.dimensiones_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.dimensiones_sheet_title,
            LinguisticStandard.DEFAULT_START_COLUMN,
            LinguisticStandard.DEFAULT_START_ROW,
            LinguisticStandard.DEFAULT_END_CELL,
        )

        if self.dimensiones_data_frame is not None:
            for r in self.dimensiones_data_frame.iterrows():
                func = MammutBratConfiguration(
                    r[1][
                        MammutBratConfiguration.DIMENSION_CLASS_GROUP_COLUMN_NAME
                    ],
                    r[1][MammutBratConfiguration.DIMENSION_CLASS_COLUMN_NAME],
                    r[1][MammutBratConfiguration.DIMENSION_COLUMN_NAME],
                    self,
                )
                self.dimensiones_descriptor_list[func.dimension_class] = func
        else:
            self.valid = False
            self.errors.append(f"The dimensiones sheet is not present.")

    def load_functionals_dataframe(self):
        # Load functional lemmas.
        self.functionals_range_name = (
                self.functionals_sheet_title
                + "!"
                + FunctionalLemmaDescriptor.DEFAULT_START_CELL
        )
        self.functionals_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.functionals_sheet_title,
            FunctionalLemmaDescriptor.DEFAULT_START_COLUMN,
            FunctionalLemmaDescriptor.DEFAULT_START_ROW,
            FunctionalLemmaDescriptor.DEFAULT_END_CELL,
        )

        if self.functionals_data_frame is not None:
            if (
                    FunctionalLemmaDescriptor.HIDE_COLUMN_NAME
                    in self.functionals_data_frame
            ):
                self.functionals_data_frame = self.functionals_data_frame[
                    self.functionals_data_frame[
                        FunctionalLemmaDescriptor.HIDE_COLUMN_NAME
                    ]
                    == ""
                    ]
            for r in self.functionals_data_frame.iterrows():
                func = FunctionalLemmaDescriptor(r[1], self)
                if func.is_explicit():
                    self.functionals[func.id] = func
                    functionals_by_lemma = {}
                    if (
                            func.regional_settings_lemma
                            in self.functionals_by_regional_settings_lemma
                    ):
                        functionals_by_lemma = self.functionals_by_regional_settings_lemma[
                            func.regional_settings_lemma
                        ]
                    else:
                        self.functionals_by_regional_settings_lemma[
                            func.regional_settings_lemma
                        ] = functionals_by_lemma
                    if func.lemma in functionals_by_lemma:
                        functionals_by_lemma[func.lemma].append(func)
                    else:
                        functionals_by_lemma[func.lemma] = [func]
                else:
                    self.functionals_implicit[
                        func.implicit_condition_id
                    ] = func
                    functionals_by_implicit_condition_id = {}
                    if (
                            func.regional_settings_lemma
                            in self.functionals_by_regional_settings_lemma_implicit
                    ):
                        functionals_by_implicit_condition_id = self.functionals_by_regional_settings_lemma_implicit[
                            func.regional_settings_lemma
                        ]
                    else:
                        self.functionals_by_regional_settings_lemma_implicit[
                            func.regional_settings_lemma
                        ] = functionals_by_implicit_condition_id
                    if (
                            func.implicit_condition_id
                            in functionals_by_implicit_condition_id
                    ):
                        functionals_by_implicit_condition_id[
                            func.implicit_condition_id
                        ].append(func)
                    else:
                        functionals_by_implicit_condition_id[
                            func.implicit_condition_id
                        ] = [func]
                if not func.valid:
                    self.valid = False
                    self.errors = self.errors + func.errors
                if func.undefined:
                    if (
                            func.regional_settings_lemma
                            in self.undefined_functionals
                    ):
                        self.undefined_functionals[
                            func.regional_settings_lemma
                        ].append(func)
                    else:
                        self.undefined_functionals[
                            func.regional_settings_lemma
                        ] = [func]
        else:
            self.valid = False
            self.errors.append(f"The functionals sheet is not present.")

    def load_non_functionals_dataframe(self):
        # Load non-functional lemmas.
        from mammut.common.lexicon.dictionary import Dictionary
        self.non_functionals_range_name = (
                self.non_functionals_sheet_title
                + "!"
                + FunctionalLemmaDescriptor.DEFAULT_START_CELL
        )
        self.non_functionals_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.non_functionals_sheet_title,
            FunctionalLemmaDescriptor.DEFAULT_START_COLUMN,
            FunctionalLemmaDescriptor.DEFAULT_START_ROW,
            NonFunctionalLemmaDescriptor.DEFAULT_END_CELL,
        )

        if self.non_functionals_data_frame is not None:
            non_functionals_data_frame_parts = []
            for i in range(0, self.non_functionals_data_frame.shape[1], 4):
                df_part = self.non_functionals_data_frame.iloc[:, i: i + 4]
                df_part = df_part.loc[
                          lambda df: df[FunctionalLemmaDescriptor.LEMMA_COLUMN_NAME]
                                     != "",
                          :,
                          ]
                non_functionals_data_frame_parts.append(df_part)
            self.non_functionals_data_frame = pd.concat(
                non_functionals_data_frame_parts
            )
            if (
                    NonFunctionalLemmaDescriptor.HIDE_COLUMN_NAME
                    in self.non_functionals_data_frame
            ):
                self.non_functionals_data_frame = self.non_functionals_data_frame[
                    self.non_functionals_data_frame[
                        NonFunctionalLemmaDescriptor.HIDE_COLUMN_NAME
                    ]
                    == ""
                    ]
            for r in self.non_functionals_data_frame.iterrows():
                nfunc = NonFunctionalLemmaDescriptor(r[1], self)
                if nfunc.regional_settings_lemma in self.non_functionals:
                    if (
                            nfunc.lemma
                            in self.non_functionals[nfunc.regional_settings_lemma]
                    ):
                        self.non_functionals[nfunc.regional_settings_lemma][
                            nfunc.lemma
                        ].append(nfunc.pos_paradigm)
                    else:
                        self.non_functionals[nfunc.regional_settings_lemma][
                            nfunc.lemma
                        ] = [nfunc.pos_paradigm]
                else:
                    self.non_functionals[nfunc.regional_settings_lemma] = {
                        nfunc.lemma: [nfunc.pos_paradigm]
                    }
        else:
            self.valid = False
            self.errors.append(f"The functionals sheet is not present.")
        self.base_dictionary = Dictionary(
            self.spreadsheet_id,
            self.primes_sheet_title,
            self.synonym_sheet_title,
            self.tokenizer,
            self,
            True,
        )

    def get_morpheme_paradigms(self, postag_value_id, regional_settings: str):
        res1 = []
        res2 = []
        if regional_settings not in self.morpheme_paradigms[postag_value_id]:
            regional_settings = ""
        if regional_settings in self.morpheme_paradigms[postag_value_id]:
            res1 = list(
                map(
                    lambda p: p.word_model,
                    filter(
                        lambda p: not p.hidden,
                        self.morpheme_paradigms[postag_value_id][
                            regional_settings
                        ].producers.values(),
                    ),
                )
            )
            res2 = list(
                map(
                    lambda p: p,
                    filter(
                        lambda p: not p.hidden,
                        self.morpheme_paradigms[postag_value_id][
                            regional_settings
                        ].producers.values(),
                    ),
                )
            )
        return res1, res2

    def get_misspelling_paradigms(
        self, postag_value_id, regional_settings: str
    ):
        res = []
        if (
            regional_settings
            not in self.misspelling_paradigms[postag_value_id]
        ):
            regional_settings = ""
        if regional_settings in self.misspelling_paradigms[postag_value_id]:
            res = list(
                map(
                    lambda p: p.word_model,
                    filter(
                        lambda p: not p.hidden,
                        self.misspelling_paradigms[postag_value_id][
                            regional_settings
                        ].producers.values(),
                    ),
                )
            )
        return res

    def get_feature_paradigms(self, postag_value_id, regional_settings: str):
        res = []
        if regional_settings not in self.feature_paradigms[postag_value_id]:
            regional_settings = ""
        if regional_settings in self.feature_paradigms[postag_value_id]:
            res = list(
                map(
                    lambda p: p.word_model,
                    filter(
                        lambda p: not p.hidden,
                        self.feature_paradigms[postag_value_id][
                            regional_settings
                        ].producers.values(),
                    ),
                )
            )
        return res

    def get_morphemes(
        self, pos_id: str, word_model: str, lemma: str, regional_settings: str
    ):
        ret = []
        lemma = lemma.lower()
        if pos_id in self.morpheme_paradigms:
            if regional_settings in self.morpheme_paradigms[pos_id]:
                pos_parad = self.morpheme_paradigms[pos_id][
                    regional_settings
                ].producers
                if word_model in pos_parad:
                    ret = pos_parad[word_model].apply(lemma)
        if len(ret) == 0:
            ret.append(
                InflectionRule.Morpheme(lemma, [], regional_settings, [], [])
            )
        return ret

    def get_misspelling(
        self, pos_id: str, word_model: str, lemma: str, regional_settings: str
    ):
        ret = []
        lemma = lemma.lower()
        if pos_id in self.misspelling_paradigms:
            pos_parad = self.misspelling_paradigms[pos_id][
                regional_settings
            ].producers
            if word_model in pos_parad:
                ret = pos_parad[word_model].apply(lemma)
        return ret

    def get_other_inflections(
        self, pos_id: str, word_model: str, lemma: str, regional_settings: str
    ):
        ret = []
        lemma = lemma.lower()
        if pos_id in self.feature_paradigms:
            pos_parad = self.feature_paradigms[pos_id][
                regional_settings
            ].producers
            if word_model in pos_parad:
                ret = pos_parad[word_model].apply(lemma)
        return ret

    def get_annotation_dimension_class(self, dimension):
        return self.dimensiones_descriptor_list[dimension]

    def get_functional_lemma_descriptor(
        self, lemma: str, regional_settings_lemma: str, pos_paradigm: str
    ):
        f_id = FunctionalLemmaDescriptor.build_id(
            lemma, regional_settings_lemma, pos_paradigm
        )
        res = None
        if f_id in self.functionals:
            res = self.functionals[f_id]
        return res

    def get_implicit_functional_lemma_descriptor(
        self,
        feature_class: str,
        feature_id: str,
        feature_pos: str,
        regional_settings_lemma: str,
    ):
        f_id = FunctionalLemmaDescriptor.build_implicit_condition_id(
            feature_class, feature_id, feature_pos, regional_settings_lemma
        )
        res = None
        if f_id in self.functionals_implicit:
            res = self.functionals_implicit[f_id]
        return res

    def has_functionals_by_lemma_regional_settings(
        self, lemma: str, regional_settings_lemma: str
    ):
        return (
            regional_settings_lemma
            in self.functionals_by_regional_settings_lemma
            and lemma
            in self.functionals_by_regional_settings_lemma[
                regional_settings_lemma
            ]
        )

    def get_functionals_lemma_for_regional_settings(
        self, regional_settings_lemma: str
    ):
        return list(
            self.functionals_by_regional_settings_lemma[
                regional_settings_lemma
            ].keys()
        )

    def get_undefined_functionals(self, regional_settings_lemma: str):
        if regional_settings_lemma in self.undefined_functionals:
            return self.undefined_functionals[regional_settings_lemma]
        else:
            return []

    def set_working_dictionary(self, working_dictionary):
        self.working_dictionary = working_dictionary

    def is_in_vocabulary(
        self, text: str, regional_settings: str, knowledge_graph=None
    ):
        tokenize_res = self.tokenizer.tokenize_out_of_stats(
            text, regional_settings, knowledge_graph
        )
        res_base = []
        res = (
            tokenize_res.unknowns
            + tokenize_res.invalid_variables
            + tokenize_res.invalid_images
            + tokenize_res.invalid_verbs
        )
        for w in tokenize_res.words:
            # A word is out of vocabulary if is not in non_functionals, not in functionals, not in the working dictionary (including the synonyms), and not in the base dictionary (including the synonyms).
            if (
                w not in self.non_functionals[regional_settings]
                and not self.has_functionals_by_lemma_regional_settings(
                    w, regional_settings
                )
                and not self.base_dictionary.contains_morpheme(w)
                and not self.base_dictionary.contains_synonym(w)
                and not self.working_dictionary.contains_morpheme(w)
                and not self.working_dictionary.contains_synonym(w)
            ):
                res.append(w)
            if self.base_dictionary.contains_morpheme(w):
                res_base.append(w)
        desc = (
            "OVR: " + str(len(res)) + "/" + str(len(tokenize_res.words))
        )  # Out of Vocabulary Ratio
        base_desc = (
            "SPR: " + str(len(res_base)) + "/" + str(len(tokenize_res.words))
        )  # Semantic Primes Ratio
        res_details = [
            "[{0} | {1:.2f}]".format(
                w,
                self.tokenizer.stats[
                    regional_settings
                ].get_word_lexical_relevance_index(w, False),
            )
            for w in res
        ]
        details = " <> ".join(res_details)
        return InVocabulary(
            len(res) == 0,
            res,
            desc,
            details,
            len(res_base) == len(tokenize_res.words),
            res_base,
            base_desc,
        )



import enum


class LanguageLevel(enum.Enum):
    """Enumeration type for language levels
    in Mammut language architecture.

    https://doc.mammut.io/docs_es/framework/language_arquitecture/
    """

    LexicalMorphological = 1
    LexicalSemantic = 2
    SyntacticSemantic = 3
    SemanticPragmatic = 4

    @staticmethod
    def from_str(label):
        if label == "LexicalMorphological":
            return LanguageLevel.LexicalMorphological
        elif label == "LexicalSemantic":
            return LanguageLevel.LexicalSemantic
        elif label == "SyntacticSemantic":
            return LanguageLevel.SyntacticSemantic
        elif label == "SemanticPragmatic":
            return LanguageLevel.SemanticPragmatic
        else:
            raise NotImplementedError


class LanguageCompetence(enum.Enum):
    """Enumeration type for language competences.

    Language competences are directly evaluated through
    Lessons in the Curriculum.

    Each Lesson will assess a language competence.
    """

    Conjugation = 1
    GenderIdentification = 2
    NumberIdentification = 3
    WordIdentification = 4
    ReadingCapabilities = 5

    @staticmethod
    def from_str(label):
        if label == "Conjugation":
            return LanguageCompetence.Conjugation
        elif label == "GenderIdentification":
            return LanguageCompetence.GenderIdentification
        elif label == "NumberIdentification":
            return LanguageCompetence.NumberIdentification
        elif label == "WordIdentification":
            return LanguageCompetence.WordIdentification
        elif label == "ReadingCapabilities":
            return LanguageCompetence.ReadingCapabilities
        else:
            raise NotImplementedError



# coding=utf-8

import tensorflow as tf
from abc import *
from mammut.models import DISTRIBUTED_TRAINING
from mammut.models.configurable import Configurable


class Node(Configurable):
    def __init__(
        self,
        scope_name,
        configuration_path,
        configuration_name,
        task_index,
        job_name,
        redis_connection_pool=None,
    ):
        Configurable.__init__(
            self, configuration_path, configuration_name, redis_connection_pool
        )
        self._scope = scope_name
        self._job = job_name
        self._task = task_index

    @abstractmethod
    def _operation(self, *args, **kwargs):
        ...

    def __call__(self, *args, **kwargs):
        if DISTRIBUTED_TRAINING:
            with tf.device("/job:{}/task:{:d}".format(self._job, self._task)):
                with tf.name_scope(self._scope):
                    return self._operation(args, kwargs)
        else:
            with tf.name_scope(self._scope):
                return self._operation(args, kwargs)



import transitions


class CurriculumStateMachineMixin(object):
    """Models the finite state machine for a concrete Curriculum
    learning experience.

    Note: This is a separate class, but some transitions are tightly coupled
    to the curriculum concrete class. This is so, because some transitions
    will call curriculum methods "on_enter"/"on_exit".
    """

    # States
    """
        Unprepared state after instantiation. 
        Call load method for transition.
    """
    STATE_UNPREPARED = "unprepared"

    """
        Ready after loaded.
    """
    STATE_READY = transitions.State(
        name="ready", on_enter="_compute_next_academic_period"
    )

    """
        When assigned, the curriculum automatically 
        will process the courses for the assigned academic period.
    """
    STATE_ASSIGNED = transitions.State(
        name="assigned", on_enter="_take_courses"
    )

    """
        Completed when all courses has been approved.
    """
    STATE_COMPLETED = "completed"

    """
        Currently, a curriculum is failed if:
            - A course can't be approved.
    """
    STATE_FAILED = "failed"

    STATES = [
        STATE_UNPREPARED,
        STATE_READY,
        STATE_ASSIGNED,
        STATE_COMPLETED,
        STATE_FAILED,
    ]

    # Transitions
    PREPARE_TRIGGER = "prepare"
    BEGIN_COURSES_TRIGGER = "begin_courses"
    ASSIGNED_COURSES_COMPLETED_TRIGGER = "assigned_courses_completed"
    CURRICULUM_COMPLETED_TRIGGER = "assigned_courses_evaluated"
    FAIL_COURSE_TRIGGER = "fail_course"

    PREPARE_TRANSITION = {
        "trigger": PREPARE_TRIGGER,
        "source": STATE_UNPREPARED,
        "dest": STATE_READY,
    }
    BEGIN_COURSES_TRANSITION = {
        "trigger": BEGIN_COURSES_TRIGGER,
        "source": STATE_READY,
        "dest": STATE_ASSIGNED,
    }
    COMPLETED_TRANSITION = {
        "trigger": CURRICULUM_COMPLETED_TRIGGER,
        "source": STATE_READY,
        "dest": STATE_COMPLETED,
    }
    COURSES_COMPLETED_TRANSITION = {
        "trigger": ASSIGNED_COURSES_COMPLETED_TRIGGER,
        "source": STATE_ASSIGNED,
        "dest": STATE_READY,
    }
    COURSE_FAILED_TRANSITION = {
        "trigger": FAIL_COURSE_TRIGGER,
        "source": STATE_READY,
        "dest": STATE_FAILED,
    }

    TRANSITIONS = [
        PREPARE_TRANSITION,
        BEGIN_COURSES_TRANSITION,
        COMPLETED_TRANSITION,
        COURSES_COMPLETED_TRANSITION,
        COURSE_FAILED_TRANSITION,
    ]

    def __init__(self):
        self.machine = transitions.Machine(
            model=self,
            states=self.STATES,
            transitions=self.TRANSITIONS,
            initial=self.STATE_UNPREPARED,
        )
        super(CurriculumStateMachineMixin, self).__init__()



# coding=utf-8
import re
from collections import namedtuple, defaultdict, UserList, UserDict

import sys

from mammut.common.basic_tokenizer import BasicTokenizer
import mammut
import logging

from brat_widget.configuration import CollectionConfiguration

from mammut.common.constants import DictionaryTable
from mammut.common.indexer import DefinitionIndexer
from mammut.common.lexicon.inflection_rule import InflectionRule
from mammut.common.util import StringEnum
import pandas as pd
from mammut.common.storage.storage_manager import StorageManager
from mammut.common.lexicon.pos import POSType

log = logging.getLogger(__name__)


class MammutBratConfiguration(CollectionConfiguration):
    DIMENSION_CLASS_GROUP_COLUMN_NAME = "dimension-class-group"
    DIMENSION_CLASS_COLUMN_NAME = "dimension-class"
    DIMENSION_COLUMN_NAME = "dimension"

    def __init__(
        self, dimension_class_group: str, dimension_class: str, dimension: str, standard
    ):
        CollectionConfiguration.__init__(self)
        self.dimension_class_group = dimension_class_group
        self.dimension_class = dimension_class
        self.dimension = dimension
        self.standard = standard
        self.pos_classes = [s.strip() for s in self.dimension.split(",") if s.strip()]
        self._arcs_for_event_types = []
        self.ui_names = {
            "entities": "pos",
            "events": "pos-arcs",
            "relations": "relations",
            "attributes": "features",
        }
        self._has_been_initialized = False

    def initialize(self):
        if not self._has_been_initialized:
            self._populate_entity_types()
            self._populate_event_types()
            CollectionConfiguration.initialize(self)
            self._has_been_initialized = True

    def _populate_event_types(self):
        for c_id in self.pos_classes:
            for pos in self.standard.pos_descriptor_classes[c_id]:
                if pos.pos_type == POSType.Event:
                    pos_fets = []
                    fet_by_class = defaultdict(list)
                    for f in self.standard.all_feature_descriptor_by_pos_tag[pos.id]:
                        fet_by_class[f.feature_class].append(f)
                    for fc, fs in fet_by_class.items():
                        att_type = f"{pos.id}-{fc}"
                        pos_fets.append(att_type)
                        a_vals = []
                        if len(fs) == 1:
                            a_vals = [{"box": "crossed", "name": fs[0].feature_id}]
                        else:
                            a_vals = []
                            for f in fs:
                                a_vals.append(
                                    {
                                        "dashArray": ",",
                                        "name": f.feature_id,
                                        "glyph": "\u2191",
                                    }
                                )
                        a_d = {
                            "unused": False,
                            "values": a_vals,
                            "labels": None,
                            "type": att_type,
                            "name": fc,
                        }
                        self.event_attribute_types.append(a_d)
                    p_d = {
                        "borderColor": "darken",
                        "normalizations": [],
                        "name": pos.name,
                        "arcs": self._arcs_for_event_types,
                        "labels": [pos.name, pos.pos_abbreviation],
                        "children": [],
                        "unused": False,
                        "bgColor": "lightgreen",
                        "attributes": pos_fets,
                        "type": pos.id,
                        "fgColor": "black",
                    }
                    p_d_arc = {
                        "labelArrow": "none",
                        "arrowHead": "triangle,5",
                        "color": "black",
                        "name": pos.name,
                        "labels": [pos.name, pos.name],
                        "dashArray": ",",
                        "type": f"{pos.id}-Arg",
                        "count": "*",  # '{1-5}', '{2}', '', '?', '*', '+'
                        "targets": [pos.id],
                    }
                    self.event_types.append(p_d)
                    self._arcs_for_event_types.append(p_d_arc)
        if self.dimension_class == "functionals":  # TODO:puede quedar hardcode??
            for key, functional in self.standard.functionals.items():
                pos_fets = []
                for definition in functional.definitions:
                    att_type = functional.lemma + "-" + definition.index  # f'elieser'
                    pos_fets.append(att_type)
                    attributes = [i.index for i in functional.definitions]
                    a_vals = []
                    for f in attributes:
                        a_vals.append({"dashArray": ",", "name": f, "glyph": "\u2191"})
                    a_d = {
                        "unused": False,
                        "values": a_vals,
                        "labels": None,
                        "type": att_type,
                        "name": "index",
                    }
                    self.event_attribute_types.append(a_d)
                p_d = {
                    "borderColor": "darken",
                    "normalizations": [],
                    "name": functional.lemma,
                    "arcs": self._arcs_for_event_types,
                    "labels": [
                        functional.lemma,
                        functional.lemma,
                    ],  # conviene que tenga abreviatura, implica agregarla al sheet
                    "children": [],
                    "unused": False,
                    "bgColor": "lightgreen",
                    "attributes": pos_fets,
                    "type": functional.id,
                    "fgColor": "black",
                }
                p_d_arc = {
                    "labelArrow": "none",
                    "arrowHead": "triangle,5",
                    "color": "black",
                    "name": functional.lemma,
                    "labels": [functional.lemma, functional.lemma],
                    "dashArray": ",",
                    "type": f"{functional.id}-Arg",
                    "count": "*",  # '{1-5}', '{2}', '', '?', '*', '+'
                    "targets": [functional.id],
                }
                self.event_types.append(p_d)
                self._arcs_for_event_types.append(p_d_arc)

    def _populate_entity_types(self):
        for c_id in self.pos_classes:
            for pos in self.standard.pos_descriptor_classes[c_id]:
                if pos.pos_type == POSType.Entity:
                    pos_fets = []
                    fet_by_class = defaultdict(list)
                    for f in self.standard.all_feature_descriptor_by_pos_tag[pos.id]:
                        fet_by_class[f.feature_class].append(f)
                    for fc, fs in fet_by_class.items():
                        att_type = f"{pos.id}-{fc}"
                        pos_fets.append(att_type)
                        a_vals = []
                        if len(fs) == 1:
                            a_vals = [{"box": "crossed", "name": fs[0].feature_id}]
                        else:
                            a_vals = []
                            for f in fs:
                                a_vals.append(
                                    {
                                        "dashArray": ",",
                                        "name": f.feature_id,
                                        "glyph": "\u2191",
                                    }
                                )
                        a_d = {
                            "unused": False,
                            "values": a_vals,
                            "labels": None,
                            "type": att_type,
                            "name": fc,
                        }
                        self.entity_attribute_types.append(a_d)
                    p_d = {
                        "borderColor": "darken",
                        "normalizations": [],
                        "name": pos.name,
                        "arcs": [],
                        "labels": [pos.name, pos.pos_abbreviation],
                        "children": [],
                        "unused": False,
                        "bgColor": "#FF821C",
                        "attributes": pos_fets,
                        "type": pos.id,
                        "fgColor": "black",
                    }
                    p_d_arc = {
                        "labelArrow": "none",
                        "arrowHead": "triangle,5",
                        "color": "black",
                        "name": pos.name,
                        "labels": [pos.arc_name, pos.arc_name],
                        "dashArray": ",",
                        "type": f"{pos.id}-Arg",
                        "count": "*",  # '{1-5}', '{2}', '', '?', '*', '+'
                        "targets": [pos.id],
                    }
                    self.entity_types.append(p_d)
                    self._arcs_for_event_types.append(p_d_arc)



# coding=utf-8
import redis

class ThoughtTuple:
    def __init__(self, user_id:int, id: int, need: str, offers: [], tokenized_need=None, tokenized_offers=None, need_target_vector=None, offers_target_vector=None):
        self.user_id = user_id
        self.id = id
        self.need = need
        self.offers = offers
        self.tokenized_need = tokenized_need
        self.tokenized_offers = tokenized_offers
        self.need_target_vector = need_target_vector
        self.offers_target_vector = offers_target_vector

    def print(self):
        print(f'\t-> ThoughtTuple: {self.id}')
        print(f'\t\tOffers:')
        for o in self.offers:
            print(f'\t\t\t{o}')
        print(f'\t\tNeed:\n\t\t\t{self.need}')


class RedisMemory:
    def __init__(self, host: str, port:int):
        self.host = host
        self.port = port
        self.r = redis.StrictRedis(host=self.host, port=self.port, db=0)

    def set_key (self, key: str, value: str, pipe=None):
        if pipe is None:
            return self.r.set(key, value)
        else:
            return pipe.set(key, value)

    def set_hash (self, key: str, items: {}, pipe=None):
        if pipe is None:
            return self.r.hmset(key, items)
        else:
            return pipe.hmset(key, items)

    def save_tuple(self, tuple: ThoughtTuple, pipe=None):
        #save hash

        tupleid = "tuple:" + str(tuple.user_id) + ":" + str(tuple.id)

        #To-Do: Switch NEED/OFFER keys
        tuple_data = {}
        tuple_data["OFFER"] = tuple.offers[0]
        tuple_data["NEED"] = tuple.need

        self.set_hash (tupleid, tuple_data, pipe)

        #save extended corpus

        for offer in tuple.offers:
            offer_message =  str(tuple.user_id) + ":" + offer
            self.set_key(offer_message, tuple.id, pipe)

    def save_thought_tuple_list(self, data):# data -> ThoughtTuple array
        #instance pipeline
        pipe = self.r.pipeline()
        for i in data:
            self.save_tuple(i, pipe)
        pipe_result = pipe.execute()

    def save_variables(self, variables):#variables ->dict
        pipe = self.r.pipeline()
        for variable in variables:
            self.set_key(variable, variables[variable], pipe)
        pipe.execute()

    def save_data(self, data, command):
        cases = {'set': self.save_variables,
                 'thoughts-hmset': self.save_thought_tuple_list #TODO: no es hmset estrictamente hablando
                 }
        try:
            cases[command](data)
        except Exception as ex:
            print(ex)



from mammut.assessment.problem import Problem
from mammut.curriculum.models.mock import TestMockModel
import time
from typing import Dict


class MockProblem(Problem):
    """Concrete class to evaluate a Mock curriculum process
    in testing.
    """

    def __init__(
        self,
        course_id: str,
        assessment_id: str,
        schedule: str,
        problem_id: str,
        problem_type: str,
        parameters: str,
        language_competence: str,
        weight: str,
        retry: str,
        observations: str,
    ):
        Problem.__init__(
            self,
            course_id,
            assessment_id,
            schedule,
            problem_id,
            problem_type,
            parameters,
            language_competence,
            weight,
            retry,
            observations,
        )
        self._sleep_sc = 3

    def solve(self, model: TestMockModel, **kwargs) -> bool:
        """Mocks solving of a problem. Always solved.
        """
        self._retries_count += 1
        time.sleep(self._sleep_sc)
        return True

    def get_result_info(self) -> Dict:
        info = {"Test info": "Nothing, just a mock."}
        return info



# coding=utf-8

from mammut.models.node import Node

PARAMETER_NODE_JOB_NAME = "ps"


class ParameterNode(Node):
    def __init__(
        self,
        scope_name,
        configuration_path,
        configuration_name,
        task_index,
        redis_connection_pool=None,
    ):
        Node.__init__(
            self,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            PARAMETER_NODE_JOB_NAME,
            redis_connection_pool,
        )



# coding=utf-8

from abc import *
from mammut.models.worker_node import WorkerNode

import tensorflow as tf


class TrainableNode(WorkerNode):
    def __init__(
        self,
        scope_name,
        configuration_path,
        configuration_name,
        task_index,
        redis_connection_pool=None,
    ):
        WorkerNode.__init__(
            self,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            redis_connection_pool,
        )

    @abstractmethod
    def _compute_loss(self, resultOp, expected_result):
        ...

    @abstractmethod
    def _optimize(sel, global_step, loss_op):
        ...

    def training_step_op(self, global_step, input, expected_result):
        """"""

        # Evaluate the model. This returns the operation asociated to this node.
        result_op = self(input)
        loss_op = self._compute_loss(result_op, expected_result)
        optimize_op, global_step = self._optimize(global_step, loss_op)

        return [optimize_op, loss_op, global_step]



# coding=utf-8
from elasticsearch import Elasticsearch, TransportError
from mammut import ES_HOSTS, CONCORDANCE_QUERY_SIZE


class NGram:
    def __init__(self):
        self.elasticsearch = Elasticsearch(ES_HOSTS)

    def get_ngram(self, index, ngram):
        try:
            docs = self.elasticsearch.search(
                index=index,
                doc_type="doc",
                body={
                    "query": {"match": {"ngram": ngram}},
                    "size": CONCORDANCE_QUERY_SIZE,
                },
            )
            s = list()
            l_hits: list() = docs["hits"]["hits"]
            for ngrams in l_hits:
                s.append(ngrams["_source"])
            return list(s)
        except TransportError as error:
            if error.error != "index_not_found_exception":
                print(str(error))
            return []

    def get_ngram_regional_settings(self, index, ngram, regional_settings):
        try:
            docs = self.elasticsearch.search(
                index=index,
                doc_type="doc",
                body={
                    "query": {
                        "bool": {
                            "must": [
                                {"match": {"ngram": ngram}},
                                {"match": {"regional_settings": regional_settings}},
                            ]
                        }
                    },
                    "size": CONCORDANCE_QUERY_SIZE,
                },
            )
            s = list()
            l_hits: list() = docs["hits"]["hits"]
            for ngrams in l_hits:
                s.append(ngrams["_source"])
            return list(s)

        except TransportError as error:
            if error.error != "index_not_found_exception":
                print(str(error))
            return []



# coding=utf-8

from mammut.curriculum.lesson import lesson_base
from mammut.common.corpus.corpus_map import CorpusMap
from mammut.curriculum.models import sentencepiece
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class SentencepieceBPELesson(lesson_base.Lesson):
    """Implements a Lesson to work with Sentencepiece
    BPE sub word model.
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            mammut_session_context: The session context from the classroom.
        """
        lesson_base.Lesson.__init__(
            self,
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )

    def _get_model_instance(self) -> sentencepiece.SentencePieceBPEModel:
        return sentencepiece.SentencePieceBPEModel(
            self.parameters_dict, self.course, self.lesson_id
        )

    def lecture(self):
        """Loads all the prepare knowledge for the lesson.

        Note: for now only loads the corpus.
        """
        self.corpus = self.get_corpus(
            self.corpus_id, self._corpus_map_reference
        )

    def practice(
        self, mammut_session_context: MammutSessionContext,
    ):
        """Process training of sentencepiece model and save the model afterward.

        Training options information at:
        https://github.com/google/sentencepiece/blob/master/doc/options.md

        Args:
            save_path: file system folder path to save the model.
        """
        self._model.train(mammut_session_context, **{"corpus": self.corpus,})
        self._model.save(mammut_session_context)


class SentencepieceNGRAMLesson(lesson_base.Lesson):
    """Implements a Lesson to work with Sentencepiece
    NGRAM subword model
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        TODO: assessments should be completed and integrated.

        Args:
            lesson_id(str): 'id' string cell value in lesson sheet.
            course(str): 'course' string cell value in lesson sheet.
            order(str): 'order' string cell value in lesson sheet.
            corpus_id(str): 'corpus_id' string string cell value in lesson sheet.

            name(str): 'name' string cell value in lesson sheet.
            language_competence(str): 'language_competence' string cell value in lesson sheet.
            mammut_session_context: The session context from the classroom.
        """
        lesson_base.Lesson.__init__(
            self,
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )

    def _get_model_instance(self) -> sentencepiece.SentencePieceNGRAMModel:
        return sentencepiece.SentencePieceNGRAMModel(
            self.parameters_dict, self.course, self.lesson_id
        )

    def lecture(self):
        """Loads all the prepare knowledge for the lesson.

        Note: for now only loads the corpus.
        """
        self.corpus = self.get_corpus(
            self.corpus_id, self._corpus_map_reference
        )

    def practice(self, mammut_session_context: MammutSessionContext):
        """Process training of sentencepiece model and save the model afterward.

        Training options information at:
        https://github.com/google/sentencepiece/blob/master/doc/options.md

        Args:
            save_path: file system folder path to save the model.
        """
        self._model.train(mammut_session_context, **{"corpus": self.corpus,})
        self._model.save(mammut_session_context)



from abc import ABC, abstractmethod
from mammut.curriculum.core.utils import LanguageCompetence
import json
from typing import Dict
from mammut.curriculum.models.model_base import ModelBase
from mammut.curriculum.report.errors.json_parsing_errors import (
    JSONProblemParsingError,
)
from mammut.curriculum.report.utils import Phase
import mammut.assessment.util as assessment_utils


class Problem(ABC):
    """A problem is the minimal unit of assessment
    to a model.

    A problem is the analogy to an exam question.
    An exam will contain any number of problems.

    The problem has a weight that contributes to the
    final course score. If a problem is approved, then
    the problem weight will be added to the course score.
    """

    def __init__(
        self,
        course_id: str,
        assessment_id: str,
        schedule: str,
        problem_id: str,
        problem_type: str,
        parameters: str,
        language_competence: str,
        weight: str,
        retry: str,
        observations: str,
    ):
        """
        Args:
            course_id(str): 'id' value in datasource retrieved data.
            assessment_id(str): 'assessment_id' value in datasource retrieved data.
            schedule(str): 'schedule' value in datasource retrieved data.
            problem_id(str): 'problem_id' value in datasource retrieved data.
            problem_type(str): 'problem_type' value in datasource retrieved data.
                Should be a string value valid for the ProblemType enumeration.
            parameters(str): 'parameters' value in datasource retrieved data.
                Should be a valid JSON string.
            language_competence(str): 'language_competence' value in datasource retrieved data.
                Should be a string value valid for the LanguageCompetence enumeration.
            weight(str): 'weight' value in datasource retrieved data.
            observations(str): 'observations' value in datasource retrieved data.
        """
        self.course_id: int = int(course_id)
        self.assessment_id: int = int(assessment_id)
        self.schedule: int = int(schedule)
        self.problem_id: int = int(problem_id)
        self.problem_type: str = problem_type
        """
            Replace line breaks from json string.
        """
        try:
            self.parameters: Dict = json.loads(parameters.replace("\n", ""))
        except json.JSONDecodeError as e:
            raise JSONProblemParsingError(
                e,
                Phase.LOAD_ASSESSMENT_DATA,
                course_id,
                assessment_id,
                problem_id,
                assessment_utils.PARAMETERS_COLUMN,
            )
        self.language_competence: LanguageCompetence = LanguageCompetence.from_str(
            language_competence
        )
        self.weight: float = float(weight)
        self.observations: str = observations

        """
            If a problem isn't solved by the model, it can be retried 
            as many times as specified.
        """
        self._retry_value: int = int(retry)
        self._retries_count: int = 0

    def is_retry_allowed(self):
        return self._retries_count < self._retry_value

    @property
    def retries_count(self):
        return self._retries_count

    @abstractmethod
    def solve(self, model: ModelBase, **kwargs) -> bool:
        """Apply the problem to a model, and assess the result.

        Args:
            model: model instance
            kwargs: any other required concrete parameter.

        Returns:
            Boolean to indicate if trained model solved the problem
                or not.
        """
        pass

    @abstractmethod
    def get_result_info(self) -> Dict:
        """Get the results information about this problem
        solving, in json format.
        """
        pass



from mammut.common.lexicon.linguistic_standard import LinguisticStandard
from typing import List
from mammut.common.synthetic.synthetic_corpus_generator.token_generator import TokenGenerator


class PrimesGenerator(TokenGenerator):
    """Semantic primes generator class. Encapsulates primes
    generation for other modules
    """

    def __init__(self, standard: LinguisticStandard):
        self.standard = standard

    # Override abstract method
    def get_tokens(
        self, pos: str = "", regional_settings: str = "es"
    ) -> List[str]:
        semantic_primes = self.standard.base_dictionary.lemmas_entries
        semantic_primes = list(
            filter(
                lambda prime: prime.regional_settings_lemma
                == regional_settings,
                semantic_primes,
            )
        )
        semantic_primes = list(map(lambda prime: prime.lemma, semantic_primes))
        return semantic_primes



from abc import ABC, abstractmethod
from typing import Optional, Dict, List
import ray
from random import randint

from mammut.curriculum.core.mammut_session_context import MammutSessionContext
from mammut.curriculum.models.model_base import ModelBase


class VerbBuilder(ABC):
    """Abstract VerbBuilder which generalizes the operations of a concrete VerbBuilder
    implementation for Verb instances created in KtT at compile time.
    """

    @abstractmethod
    def get_builder_name(self) -> str:
        """Returns the name of the concrete builder.

        Useful for testing the builder value in real KtT process running in the
        Curriculum processing.
        """
        pass

    @abstractmethod
    def create_verb(
        self,
        title: str,
        micro_experience_type: str,
        parameter_1: str,
        parameter_2: Optional[str],
    ) -> Optional[int]:
        """Creates a Verb and save it as defined by the concrete implementation.

        Returns:
            ID of the created Verb. None if creation of the verb failed.
        """
        pass


@ray.remote
class VerbBuilderMock(VerbBuilder):
    """This VerbBuilderMock is a dummy representation of a VerbBuilder.

    This VerbBuilderMock is useful for testing the functionality with a real KtT
    process running in the Curriculum processing.
    """

    def get_builder_name(self) -> str:
        return "VerbBuilderMock"

    def create_verb(
        self,
        title: str,
        micro_experience_type: str,
        parameter_1: str,
        parameter_2: Optional[str],
    ) -> Optional[int]:
        """Mocks the creations of a Verb and returns it's ID.

        This is useful for testing the return value in the KtT process.

        Returns:
            The ID supposedly created for the Verb.
        """
        return randint(0, 5000)


class UIEventBuilder(ABC):
    """Abstract UIEventBuilder which generalizes the operations of a UIEvent builder implementation
    for UI Events verb instances created in KtT at compile time"""

    @abstractmethod
    def get_builder_name(self) -> str:
        """Returns the name of the concrete builder.

        Useful for testing the builder value in real KtT process running in the
        Curriculum processing.
        """
        pass

    @abstractmethod
    def create_ui_event(
        self,
        uri: str,
        control: str,
        kind: str,
        arguments: Optional[List[str]],
        text: str,
    ) -> Optional[int]:
        """Creates a UIEvent and saves it as defined by the concrete implementation.

        Returns:
             ID of the created UI Event. None if the creation of UI Event failed.
        """
        pass


@ray.remote
class UIEventBuilderMock(UIEventBuilder):
    """This UIEventBuilderMock is a dummy implementation of a UIEvent Builder.

    This is iseful for testing the functionality with a real KtT process running
    in the Curriculum processing.
    """

    def get_builder_name(self) -> str:
        return "UIEVentBuilderMock"

    def create_ui_event(
        self,
        uri: str,
        control: str,
        kind: str,
        arguments: Optional[List[str]],
        text: str,
    ) -> Optional[int]:
        """Mocks the creation of a UI Event and returns it's ID.

        Returns:
            The ID supposedly created for the UI Event.
        """
        return randint(0, 5000)


@ray.remote
class KttInterpretationModel(ModelBase):
    """This KttInterpretationModel is designed to perform the package
    compilation-interpretation process.

    Internally is composed of a ray actor handle for the KtT ray java actor.
    """

    MODEL_NAME = "KttInterpretationModel"

    def __init__(
        self, parameters: Dict, course_id: int, lesson_id: int,
    ):
        # Don't use `super` syntax. It wil cause an error with Ray.
        # https://github.com/ray-project/ray/issues/449#issuecomment-727266380
        ModelBase.__init__(self, parameters, course_id, lesson_id)
        # TODO: This string should be on the configuration files.
        self._ktt_actor_class = ray.java_actor_class(
            "com.mammut.interpreter.raycurriculum.InterpreterMainRayCurriculumService"
        )
        self._ktt_actor_handle: ray.actor.ActorHandle = None
        self._verb_builder: ray.actor.ActorHandle = None
        self._ui_event_builder: ray.actor.ActorHandle = None

    def train(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Train the model this class implements.

        Override this method in child classes. It's defined to `pass`
        here to avoid errors with Ray actor definition.

        Args:
            mammut_session_context: Mammut session context with general information about
                current curriculum instance.
            **kwargs: Named parameters required for the implementation
                in the child class.
        """

        # Creates Verb and UIEvent builders mock. Later the decision of which concrete class to instantiate
        # could be set as a configuration parameter.
        self._verb_builder = VerbBuilderMock.remote()
        self._ui_event_builder = UIEventBuilderMock.remote()

        self._ktt_actor_handle = self._ktt_actor_class.remote(
            "KtT package interpretation",
            mammut_session_context.main_report_actor_handle,
            self._verb_builder,
            self._ui_event_builder,
        )
        # Todo: Mammut-ID needs to be defined. It'll be when KtT mammut-api operations
        #   get decoupled. (https://github.com/mammut-io/mammut-py/issues/150)
        self._ktt_actor_handle.startPackageCompilation.remote(
            mammut_session_context.package_id,
            8264,
            "KtT package interpretation",
        )

    def save(self, mammut_session_context: MammutSessionContext, **kwargs):
        """The model itself has nothing to save.

        The lesson that holds this model are responsible for registering it
        in the OnlineModelsRegistry.

        Args:
            mammut_session_context: Mammut session context with general information about
                current curriculum instance.
            **kwargs: Any other custom attribute.
        """
        pass



from mammut.curriculum.report.reporters.base_reporter import BaseReporter
from mammut.curriculum.report.messages.message import CurriculumBaseMessage
from mammut.curriculum.report.reporters.console_reporter import ConsoleReporter
import logging
from typing import List

log = logging.getLogger(__name__)


class RemoteClientReporter(BaseReporter):
    """Reporter class to support client messages reporting
    in remote classrooms.

    The client can enable module logging to also publish
    messages in the conventional way.

    This remote client reporter holds an actor handle
    for a ray actor where the messages are centralized.

    TODO: constructor parameters should be replaced by
        configuration entries.
    """

    def __init__(
        self, threshold=logging.INFO, enable_module_logging: bool = False,
    ):
        """
        Args:
            threshold: threshold level for this reporter.
            enable_module_logging (bool): enable store of messages in this reporter state.
        """
        super(RemoteClientReporter, self).__init__(threshold)
        self._module_logging_enable: bool = enable_module_logging
        self._reporters: List[BaseReporter] = []
        self._reporters.append(
            ConsoleReporter("client_console_remote_reporter", self.threshold)
        )
        log.debug("Remote client reporter configured")

    def send_message(self, message: CurriculumBaseMessage, **kwargs):
        """Send the message to the remote actor handle.

        Args:
            message: Message to be sent to the remote actor handle.
            **main_report_handle: the handle of the main reporter in the main
                process.
        """
        main_report_handle = kwargs["main_report_handle"]

        if self._module_logging_enable:
            log.log(message.level, str(message))

        main_report_handle.save_message.remote(message)

        for reporter in self._reporters:
            reporter.send_message(message)
